@@ -108,6 +108,11 @@ class WaveFormat(Enum):
     A_LAW = 6
     MU_LAW = 7
     EXTENDED = 65534
+    UNSUPPORTED = -1
+
+    @classmethod
+    def has_value(cls, value):
+        return value in cls._value2member_map_
 
 
 class FormatChunk(Chunk):
@@ -116,9 +121,11 @@ class FormatChunk(Chunk):
     """
 
     __format: WaveFormat
-    __extended: bool
+    __extension: bytes
     __channels: int
     __sample_rate: int
+    __bytes_per_sec: int
+    __block_align: int
     __bits_per_sample: int
 
     LENGTH_CHUNK = 24
@@ -128,10 +135,12 @@ class FormatChunk(Chunk):
     def __init__(
         self,
         wave_format: WaveFormat,
-        extended: bool,
+        extension: bytes,
         channels: int,
         sample_rate: int,
         bits_per_sample: int,
+        bytes_per_sec: int = None,
+        block_align: int = None,
     ):
         """
         Creates a new instance of the format block.
@@ -142,13 +151,28 @@ class FormatChunk(Chunk):
             channels (int): The number of channels in the file.
             sample_rate (int): The sample rate.
             bits_per_sample (int): The number of bits in each sample.
+            bytes_per_sec (int): Average bytes per second. Computed from sample_rate, channels and bits_per_sample if None
+            block_align (int): Block alignment for the file. Computed from channels and bits_per_sample if None
+
         """
 
-        self.__format = wave_format
-        self.__extended = extended
+        if isinstance(wave_format, WaveFormat):
+            self.__format = wave_format.value
+        else:
+            self.__format = wave_format
+        self.__extension = extension
         self.__channels = channels
         self.__sample_rate = sample_rate
         self.__bits_per_sample = bits_per_sample
+        self.__bytes_per_sec = bytes_per_sec
+        if bytes_per_sec == None:
+            self.__bytes_per_sec = sample_rate * channels * bits_per_sample // 8
+        else:
+            self.__bytes_per_sec = bytes_per_sec
+        if block_align == None:
+            self.__block_align = channels * bits_per_sample // 8
+        else:
+            self.__block_align = block_align
 
     @classmethod
     def from_file(cls, file_handle: BinaryIO, offset: int) -> FormatChunk:
@@ -166,7 +190,14 @@ class FormatChunk(Chunk):
 
         # Read from the chunk
 
-        (handle, channels, sample_rate, _, _, bits_per_sample,) = unpack(
+        (
+            wave_format,
+            channels,
+            sample_rate,
+            bytes_per_sec,
+            block_align,
+            bits_per_sample,
+        ) = unpack(
             "<HHIIHH",
             seek_and_read(
                 file_handle,
@@ -175,14 +206,29 @@ class FormatChunk(Chunk):
             ),
         )
 
-        # Read the format
+        # Read extension
 
-        wave_format = WaveFormat(handle)
+        extension = None
+        if extended:
+            (extension_size,) = unpack(
+                "<H",
+                file_handle.read(2),
+            )
+            if extension_size:
+                extension = file_handle.read(extension_size)
+            else:
+                extension = b""
 
         # Generate our object
 
         return FormatChunk(
-            wave_format, extended, channels, sample_rate, bits_per_sample
+            wave_format,
+            extension,
+            channels,
+            sample_rate,
+            bits_per_sample,
+            bytes_per_sec,
+            block_align,
         )
 
     @property
@@ -190,7 +236,9 @@ class FormatChunk(Chunk):
         """
         Indicates the format the audio is encoded in.
         """
-        return self.__format
+        if WaveFormat.has_value(self.__format):
+            return WaveFormat(self.__format)
+        return f"{self.__format} {WaveFormat.UNSUPPORTED}"
 
     @property
     def channels(self) -> int:
@@ -218,21 +266,28 @@ class FormatChunk(Chunk):
         """
         The bytes per second this file is encoded at.
         """
-        return self.sample_rate * self.channels * self.bits_per_sample // 8
+        return self.__bytes_per_sec
 
     @property
     def block_align(self) -> int:
         """
         The block alignment for the file.
         """
-        return self.channels * self.bits_per_sample // 8
+        return self.__block_align
 
     @property
     def extended(self) -> bool:
         """
         Indicates if the header is extended or not.
         """
-        return self.__extended
+        return self.__extension != None
+
+    @property
+    def extension(self) -> bytes:
+        """
+        Raw extension.
+        """
+        return self.__extension
 
     @property
     def get_name(self) -> str:
@@ -240,27 +295,37 @@ class FormatChunk(Chunk):
 
     def to_bytes(self) -> List[bytes]:
 
-        # Sanity check
-
-        if self.extended:
-            raise ExportExtendedFormatException(
-                "We don't support converting extended format headers to binary blobs."
-            )
-
         # Build up our chunk
 
-        return pack(
+        length = (
+            self.LENGTH_STANDARD_SIZE
+            if self.__extension == None
+            else self.LENGTH_STANDARD_SIZE + 2 + len(self.__extension)
+        )
+
+        format = pack(
             "<4sIHHIIHH",
             self.HEADER_FORMAT,
-            self.LENGTH_STANDARD_SIZE,
-            self.format.value,
-            self.channels,
-            self.sample_rate,
-            self.byte_rate,
-            self.block_align,
-            self.bits_per_sample,
+            length,
+            self.__format,
+            self.__channels,
+            self.__sample_rate,
+            self.__bytes_per_sec,
+            self.__block_align,
+            self.__bits_per_sample,
         )
 
+        if self.__extension == None:
+            return format
+
+        # build extension size
+
+        ext_size = pack("<H", len(self.__extension))
+
+        # return complete chunk
+        data = b"".join([format, ext_size, self.__extension])
+        return data
+
 
 class DataChunk(Chunk):
     """
@@ -319,14 +384,27 @@ class DataChunk(Chunk):
         raw = file_handle.read(length)
 
         # Create the object
+        #
+
+        if wave_format.bits_per_sample // 8 == 0:
+            # unsupported data format : use raw datas
+            return DataChunk(raw)
 
         sample_count = (
             length // wave_format.channels // (wave_format.bits_per_sample // 8)
         )
-        samples = np.frombuffer(
-            raw,
-            dtype=np.dtype(f"<i{wave_format.bits_per_sample // 8}"),
-        ).reshape(sample_count, wave_format.channels)
+        if wave_format.bits_per_sample // 8 == 3:
+            # use raw data of 3 bytes as sample for 24 bits
+            # [disadvantage: the samples must be converted by caller before processing]
+            samples = np.frombuffer(
+                raw,
+                dtype=np.dtype("V3"),
+            )
+        else:
+            samples = np.frombuffer(
+                raw,
+                dtype=np.dtype(f"<i{wave_format.bits_per_sample // 8}"),
+            ).reshape(sample_count, wave_format.channels)
 
         return DataChunk(samples)
 
@@ -341,7 +419,10 @@ class DataChunk(Chunk):
 
         # Generate the data section
 
-        data = self.__samples.tobytes()
+        if isinstance(self.__samples, bytes):
+            data = self.__samples
+        else:
+            data = self.__samples.tobytes()
 
         # Generate the header
 
@@ -962,7 +1043,6 @@ class LabelChunk(Chunk):
         return self.__label
 
 
-
 class NoteChunk(Chunk):
     """
     A note associated with a cue point
@@ -1032,7 +1112,6 @@ class NoteChunk(Chunk):
         return self.__note
 
 
-
 class LabeledTextChunk(Chunk):
     """
     A labeled text
@@ -1051,7 +1130,17 @@ class LabeledTextChunk(Chunk):
     __codepage: int
     __label: str
 
-    def __init__(self, id: int, sample_length: int, purpose: str, country: int, language: int, dialect: int, codepage: int, label: str):
+    def __init__(
+        self,
+        id: int,
+        sample_length: int,
+        purpose: str,
+        country: int,
+        language: int,
+        dialect: int,
+        codepage: int,
+        label: str,
+    ):
         self.__id = id
         self.__sample_length = sample_length
         self.__purpose = purpose
@@ -1059,8 +1148,7 @@ class LabeledTextChunk(Chunk):
         self.__language = language
         self.__dialect = dialect
         self.__codepage = codepage
-        self.__label = label if label else ''
-
+        self.__label = label if label else ""
 
     @property
     def get_name(self) -> str:
@@ -1095,7 +1183,16 @@ class LabeledTextChunk(Chunk):
             )  # skipcq: TCV-001
 
         # Read the rest of the header
-        new_id, sample_length, purpose, country, language, dialect, codepage, raw_label = unpack(
+        (
+            new_id,
+            sample_length,
+            purpose,
+            country,
+            language,
+            dialect,
+            codepage,
+            raw_label,
+        ) = unpack(
             f"<II4sHHHH{length - cls.OFFSET_LABEL}s",
             seek_and_read(
                 file_handle,
@@ -1103,7 +1200,16 @@ class LabeledTextChunk(Chunk):
                 length,
             ),
         )
-        return LabeledTextChunk(new_id, sample_length, purpose, country, language, dialect, codepage, decode_string(raw_label))
+        return LabeledTextChunk(
+            new_id,
+            sample_length,
+            purpose,
+            country,
+            language,
+            dialect,
+            codepage,
+            decode_string(raw_label),
+        )
 
     @property
     def id(self) -> str:
@@ -1168,18 +1274,22 @@ class ListChunk(Chunk):
     """
 
     __sub_chunks: List[Chunk] = []
+    __list_type: str
 
     HEADER_LIST = b"LIST"
+    LENGTH_LIST_TYPE = 4
     HEADER_ASSOC = b"adtl"
+    HEADER_INFO = b"INFO"
     LENGTH_ASSOC = 4
-    OFFSET_SUBCHUNKS = 12
+    OFFSET_SUBCHUNKS = 4
     CHUNK_HEADER_MAP = {
         b"labl": LabelChunk,
         b"note": NoteChunk,
         b"ltxt": LabeledTextChunk,
     }
 
-    def __init__(self, sub_chunks: List[Chunk]):
+    def __init__(self, list_type, sub_chunks: List[Chunk]):
+        self.__list_type = list_type
         self.__sub_chunks = sub_chunks
 
     @classmethod
@@ -1193,6 +1303,13 @@ class ListChunk(Chunk):
                 "List header must start with list"
             )  # skipcq: TCV-001
 
+        # Read List type
+
+        offset += Chunk.OFFSET_CHUNK_CONTENT
+        (list_type,) = unpack(
+            "<4s", seek_and_read(file_handle, offset, ListChunk.LENGTH_LIST_TYPE)
+        )
+
         # Read in the sub chunks
 
         current_offset = offset + cls.OFFSET_SUBCHUNKS
@@ -1209,9 +1326,23 @@ class ListChunk(Chunk):
                 current_sub_chunk = chunk_type.from_file(file_handle, current_offset)
                 sub_chunks.append(current_sub_chunk)
 
-            current_offset += current_length + cls.OFFSET_CHUNK_CONTENT
+            else:
+                # create generic Chunk for holding unsupported header and datas
+                if list_type == b"INFO":
+                    current_sub_chunk = InfoChunk.from_file(
+                        file_handle, current_offset, current_header, current_length
+                    )
+                else:
+                    current_sub_chunk = GenericChunk.from_file(
+                        file_handle, current_offset, current_header, current_length
+                    )
+                sub_chunks.append(current_sub_chunk)
 
-        return ListChunk(sub_chunks)
+            current_offset += (
+                current_length + cls.OFFSET_CHUNK_CONTENT + current_length % 2
+            )
+
+        return ListChunk(list_type, sub_chunks)
 
     def to_bytes(self) -> List[bytes]:
         encoded_sub_chunks = [sub_chunk.to_bytes() for sub_chunk in self.sub_chunks]
@@ -1219,10 +1350,34 @@ class ListChunk(Chunk):
         length = len(combined_sub_chunks)
 
         header = pack(
-            "<4sI4s", self.HEADER_LIST, length + self.LENGTH_ASSOC, self.HEADER_ASSOC
+            "<4sI4s", self.HEADER_LIST, length + self.LENGTH_ASSOC, self.__list_type
         )
         return b"".join([header, combined_sub_chunks])
 
+    def get_chunk(self, chunk_name: str) -> Chunk:
+        """
+        Get sub chunk of type name
+        """
+
+        if isinstance(chunk_name, str):
+            chunk_name = str.encode(chunk_name)
+
+        for chunk in self.__sub_chunks:
+            if chunk.get_name == chunk_name:
+                return chunk
+
+        return None
+
+    def replace_chunk(self, chunk: InfoChunk):
+        """
+        replace or append a info chunk id
+        """
+        for num, _chunk in enumerate(self.__sub_chunks):
+            if chunk.get_name == _chunk.get_name:
+                self.__sub_chunks[num] = chunk
+                return
+        self.__sub_chunks.append(chunk)
+
     @property
     def sub_chunks(self) -> List[Chunk]:
         """
@@ -1234,6 +1389,10 @@ class ListChunk(Chunk):
     def get_name(self) -> str:
         return self.HEADER_LIST  # skipcq: TCV-001
 
+    @property
+    def get_type(self) -> str:
+        return self.__list_type
+
 
 class CuePoint(Chunk):
     """
@@ -1400,12 +1559,131 @@ class CueChunk(Chunk):
         return b"".join([header, *encoded_cue_points])
 
 
+class GenericChunk(Chunk):
+    """
+    Generic class for handling chunk not supported.
+    """
+
+    def __init__(self, header: bytes, datas: np.ndarray[np.int]):
+        """
+        Creates a new instance of the generic chunk.
+
+        Args:
+            datas (np.ndarray[np.int]): The data to work with.
+        """
+        self.__header = header
+        self.__datas = datas
+
+    @classmethod
+    def from_file(
+        cls, file_handle: BinaryIO, offset: int, header: bytes, length: int
+    ) -> Chunk:
+        """
+        Reads the unsupported data chunk from a file
+        """
+        (header_str, length) = cls.read_header(file_handle, offset)
+
+        # Read in the raw data
+
+        raw = file_handle.read(length)
+
+        # Create the object
+
+        datas = np.frombuffer(raw, dtype=np.dtype(np.uint8))
+
+        return GenericChunk(header_str, datas)
+
+    @property
+    def datas(self) -> np.ndarray[np.uint8]:
+        """
+        The audio sample vectors.
+        """
+        return self.__datas
+
+    def to_bytes(self) -> List[bytes]:
+
+        # Generate the data section
+
+        data = self.__datas.tobytes()
+
+        # Generate the header
+
+        header = pack("<4sI", self.__header, len(data))
+
+        # Splatter it together
+
+        return b"".join([header, data])
+
+    @property
+    def get_name(self) -> str:
+        return self.__header
+
+
+class InfoChunk(Chunk):
+    """
+    simple string chunk for LIST INFO
+    """
+
+    def __init__(self, header: bytes, info: bytes):
+        """
+        Creates a new instance of chunk.
+        """
+        if isinstance(header, str):
+            header = encode_string(header)
+        self.__header = header
+        if isinstance(info, str):
+            info = encode_string(info)
+        self.__info = info
+
+    @classmethod
+    def from_file(
+        cls, file_handle: BinaryIO, offset: int, header: bytes, length: int
+    ) -> Chunk:
+        """
+        Reads data chunk from a file
+        """
+        (header_str, length) = cls.read_header(file_handle, offset)
+
+        # Read in the raw string
+
+        (raw_info,) = unpack(f"<{length}s", file_handle.read(length))
+
+        # Create instance
+
+        return InfoChunk(header_str, decode_string(raw_info))
+
+    def to_bytes(self) -> List[bytes]:
+
+        # Generate raw string
+
+        encoded_info = null_terminate(self.__info, True)
+        encoded_info_length = len(encoded_info)
+
+        return pack(
+            f"<4sI{encoded_info_length}s",
+            self.__header,
+            encoded_info_length,
+            encoded_info,
+        )
+
+    @property
+    def info(self) -> str:
+        """
+        The info string.
+        """
+        return self.__info
+
+    @property
+    def get_name(self) -> str:
+        return self.__header
+
+
 class RiffChunk(Chunk):
     """
     The second level WAVE chunk in a RIFF file.
     """
 
-    __sub_chunks: Dict[str, Chunk] = {}
+    __sub_chunks: List[Chunk] = []
 
     HEADER_RIFF = b"RIFF"
     HEADER_WAVE = b"WAVE"
@@ -1431,7 +1709,7 @@ class RiffChunk(Chunk):
     STRUCT_SUB_TYPE = "<4s"
     STRUCT_RIFF_HEADER = "<4sI4s"
 
-    def __init__(self, sub_chunks: Dict[str, Chunk]) -> None:
+    def __init__(self, sub_chunks: List[Chunk]) -> None:
         self.__sub_chunks = sub_chunks
 
     @classmethod
@@ -1465,7 +1743,7 @@ class RiffChunk(Chunk):
         current_offset = offset + cls.OFFSET_CHUNKS_START
         file_handle.seek(0, 2)
         end_of_file = min(file_handle.tell(), length + current_offset)
-        chunk_map = {}
+        chunk_list = []
 
         while current_offset < end_of_file:
             (current_header, current_length) = cls.read_header(
@@ -1475,23 +1753,38 @@ class RiffChunk(Chunk):
 
             if chunk_type:
                 if chunk_type == DataChunk:
-                    if b"fmt " not in chunk_map:
+                    chunk_format = None
+                    for chunk in chunk_list:
+                        if chunk.get_name == b"fmt ":
+                            chunk_format = chunk
+                            break
+                    if not chunk_format:
                         raise InvalidWaveException(
                             "A format chunk must be read before a data chunk!"
                         )
                     chunk = DataChunk.from_file_with_format(
-                        file_handle, current_offset, chunk_map.get(cls.CHUNK_FORMAT)
+                        file_handle, current_offset, chunk_format
                     )
                 else:
                     chunk = chunk_type.from_file(file_handle, current_offset)
 
-                chunk_map[current_header] = chunk
+                chunk_list.append(chunk)
+
+            else:
+                # create generic Chunk for holding unsupported header and datas
+                chunk = GenericChunk.from_file(
+                    file_handle, current_offset, current_header, current_length
+                )
+
+                chunk_list.append(chunk)
 
             # Cycle onto the next chunk
 
-            current_offset += current_length + cls.OFFSET_CHUNK_CONTENT
+            current_offset += (
+                current_length + cls.OFFSET_CHUNK_CONTENT + current_length % 2
+            )
 
-        return RiffChunk(chunk_map)
+        return RiffChunk(chunk_list)
 
     @property
     def sub_chunks(self) -> Dict[str, Chunk]:
@@ -1505,27 +1798,29 @@ class RiffChunk(Chunk):
 
         # Check we have at least a format and data chunk
 
-        if self.CHUNK_FORMAT not in self.sub_chunks:
+        if not self.get_chunk(self.CHUNK_FORMAT):
             raise InvalidWaveException("Valid wave files must have a format chunk")
 
-        if self.CHUNK_DATA not in self.sub_chunks:
+        if not self.get_chunk(self.CHUNK_DATA):
             raise InvalidWaveException("Valid wave files must have a data chunk")
 
-        # Build our chunks up in order
+        # Build our chunks
 
+        #  There are no restrictions upon the order of the chunks, except :
+        #       the Format chunk must precede the Data chunk. Some inflexibly written programs expect the Format chunk as the first chunk (after the RIFF header)
+        #  (source: http://midi.teragonaudio.com/tech/wave.htm)
         chunk_bytes = []
-        chunk_bytes.append(self.sub_chunks.get(self.CHUNK_FORMAT).to_bytes())
+        chunk_bytes.append(self.get_chunk(self.CHUNK_FORMAT).to_bytes())
 
-        if self.CHUNK_CART in self.sub_chunks:
-            chunk_bytes.append(self.sub_chunks.get(self.CHUNK_CART).to_bytes())
+        # Build all other chunks
 
-        if self.CHUNK_LIST in self.sub_chunks:
-            chunk_bytes.append(self.sub_chunks.get(self.CHUNK_LIST).to_bytes())
+        for chunk in self.sub_chunks:
 
-        if self.CHUNK_CUE in self.sub_chunks:
-            chunk_bytes.append(self.sub_chunks.get(self.CHUNK_CUE).to_bytes())
+            if chunk.get_name == self.CHUNK_FORMAT:
+                # already done
+                continue
 
-        chunk_bytes.append(self.sub_chunks.get(self.CHUNK_DATA).to_bytes())
+            chunk_bytes.append(chunk.to_bytes())
 
         # Create the header
 
@@ -1539,3 +1834,43 @@ class RiffChunk(Chunk):
         # Create a blob
 
         return b"".join(chunk_bytes)
+
+    def get_chunks(self, chunk_type: str) -> Chunk:
+        """
+        Get all chunks of type name
+        """
+
+        if isinstance(chunk_type, str):
+            chunk_type = str.encode(chunk_type)
+
+        return [chunk for chunk in self.__sub_chunks if chunk.get_name == chunk_type]
+
+    def get_chunk(self, chunk_name: str, chunk_type=None, index=0) -> Chunk:
+        """
+        Get chunk by type name (and sub type for chunk LIST)
+        """
+
+        chunks = self.get_chunks(chunk_name)
+        if chunk_type:
+            for chunk in chunks:
+                if isinstance(chunk, ListChunk) and chunk.get_type == chunk_type:
+                    return chunk
+            return None
+
+        if index < len(chunks):
+            return chunks[index]
+
+        return None
+
+    def replace_chunk(self, chunk: Chunk):
+        """
+        replace or append a chunk id
+        """
+        for num, _chunk in enumerate(self.__sub_chunks):
+            if chunk.get_name == _chunk.get_name:
+                if isinstance(chunk, ListChunk):
+                    if chunk.get_type != _chunk.get_type:
+                        continue
+                self.__sub_chunks[num] = chunk
+                return
+        self.__sub_chunks.append(chunk)
